import { Builtin } from "ts-essentials/dist/types";
import { NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { DeepOmit, StrictOmit } from "ts-essentials";

export function assertUnreachable(x: never): never {
  throw new Error(`Didn't expect to get here: ${JSON.stringify(x as any)}`);
}

export function assertUnreachableNoError(_x: never): void {}

export function isDefined<T>(object: T | undefined): object is T {
  return object !== undefined;
}

export function isDefinedEntry<T, K extends string>(object: [K, T | undefined]): object is [K, T] {
  return object[1] !== undefined;
}

export function enumerateTypes<K extends keyof any>(record: Record<K, null>): K[] {
  return Object.keys(record) as any;
}

export function optional<T>(item: T | undefined): T[] {
  return item === undefined ? [] : [item];
}

export function deepNever<T extends DeepOmitModify<Filter>, Filter>(
  item: DeepOmit<T, Filter>
): T & { meta: { rollbackState: never } } {
  return item as any;
}

/**
 * An endofunction is a function whose domain is equal to its range (categorically / statically speaking).
 */
export type Endo<T> = (item: T) => T;

export type Reduce<T> = (left: T, right: T) => T;

export type Predicate<T> = (item: T) => boolean;

/**
 * An arbitrary serializable primitive.
 */
export declare type JsonPrimitive = string | number | boolean | null; // 'undefined' is not serializable.

/**
 * An arbitrary JSON-serializable complex object.
 *
 * Note: arrays must be one-dimensional. This is due to a limitation in our tooling (TypeScript and ts-essentials) that
 * requires recursive types to be expressed as a single type rather than a series of cyclic types, thus precluding
 * arrays of an unknown dimension: if N-dimensional arrays are required, we can add support for the, but only up to a
 * fixed number of dimensions (e.g. 1D, 2D...5D but not ND).
 *
 * @additionalProperties true
 */
export interface JsonObject {
  [key: string]: JsonPrimitive | JsonObject | Array<JsonPrimitive | JsonObject>;
}

export type AssertEqualKeys<T1 extends object, T2 extends object> = [
  keyof T1 extends keyof T2 ? 1 : 0,
  keyof T2 extends keyof T1 ? 1 : 0
] extends [1, 1]
  ? true
  : false;

export function assertEqualKeys<T1 extends object, T2 extends object>(_areEqual: AssertEqualKeys<T1, T2>): void {}

export type UndefinedToNull<T> = T extends undefined ? null : T;

export type DeepUndefinedToNull<T> = T extends Builtin
  ? UndefinedToNull<T>
  : T extends Map<infer K, infer V>
  ? Map<DeepUndefinedToNull<K>, DeepUndefinedToNull<V>>
  : T extends ReadonlyMap<infer K, infer V>
  ? ReadonlyMap<DeepUndefinedToNull<K>, DeepUndefinedToNull<V>>
  : T extends WeakMap<infer K, infer V>
  ? WeakMap<DeepUndefinedToNull<K>, DeepUndefinedToNull<V>>
  : T extends Set<infer U>
  ? Set<DeepUndefinedToNull<U>>
  : T extends ReadonlySet<infer U>
  ? ReadonlySet<DeepUndefinedToNull<U>>
  : T extends WeakSet<infer U>
  ? WeakSet<DeepUndefinedToNull<U>>
  : T extends Promise<infer U>
  ? Promise<DeepUndefinedToNull<U>>
  : T extends {}
  ? {
      [K in keyof T]: DeepUndefinedToNull<T[K]>;
    }
  : UndefinedToNull<T>;

export type NullToUndefined<T> = T extends null ? undefined : T;

export type DeepNullToUndefined<T> = T extends Builtin
  ? NullToUndefined<T>
  : T extends Map<infer K, infer V>
  ? Map<DeepNullToUndefined<K>, DeepNullToUndefined<V>>
  : T extends ReadonlyMap<infer K, infer V>
  ? ReadonlyMap<DeepNullToUndefined<K>, DeepNullToUndefined<V>>
  : T extends WeakMap<infer K, infer V>
  ? WeakMap<DeepNullToUndefined<K>, DeepNullToUndefined<V>>
  : T extends Set<infer U>
  ? Set<DeepNullToUndefined<U>>
  : T extends ReadonlySet<infer U>
  ? ReadonlySet<DeepNullToUndefined<U>>
  : T extends WeakSet<infer U>
  ? WeakSet<DeepNullToUndefined<U>>
  : T extends Promise<infer U>
  ? Promise<DeepNullToUndefined<U>>
  : T extends {}
  ? {
      [K in keyof T]: DeepNullToUndefined<T[K]>;
    }
  : NullToUndefined<T>;

export type EmptiableArray<T> = T extends NonEmptyArray<infer U> ? U[] : T;

export type DeepEmptiableArray<T> = T extends Builtin
  ? EmptiableArray<T>
  : T extends Map<infer K, infer V>
  ? Map<K, DeepEmptiableArray<V>>
  : T extends ReadonlyMap<infer K, infer V>
  ? ReadonlyMap<K, DeepEmptiableArray<V>>
  : T extends WeakMap<infer K, infer V>
  ? WeakMap<K, DeepEmptiableArray<V>>
  : T extends Set<infer U>
  ? Set<DeepEmptiableArray<U>>
  : T extends ReadonlySet<infer U>
  ? ReadonlySet<DeepEmptiableArray<U>>
  : T extends WeakSet<infer U>
  ? WeakSet<DeepEmptiableArray<U>>
  : T extends Promise<infer U>
  ? Promise<DeepEmptiableArray<U>>
  : T extends NonEmptyArray<infer U> // Array extends {}, so we must treat specially here (and before the {} below!).
  ? Array<DeepEmptiableArray<U>>
  : T extends {}
  ? {
      [K in keyof T]: DeepEmptiableArray<T[K]>;
    }
  : EmptiableArray<T>;

// Inverse of Opaque<T>
export type Transparent<T> = T extends { readonly __TYPE__: any } ? StrictOmit<T, "__TYPE__"> : T;

export type DeepTransparent<T> = T extends Builtin
  ? Transparent<T>
  : T extends Map<infer K, infer V>
  ? Map<K, DeepTransparent<V>>
  : T extends ReadonlyMap<infer K, infer V>
  ? ReadonlyMap<K, DeepTransparent<V>>
  : T extends WeakMap<infer K, infer V>
  ? WeakMap<K, DeepTransparent<V>>
  : T extends Set<infer U>
  ? Set<DeepTransparent<U>>
  : T extends ReadonlySet<infer U>
  ? ReadonlySet<DeepTransparent<U>>
  : T extends WeakSet<infer U>
  ? WeakSet<DeepTransparent<U>>
  : T extends Promise<infer U>
  ? Promise<DeepTransparent<U>>
  : T extends {}
  ? {
      [K in keyof T]: DeepTransparent<T[K]>;
    }
  : Transparent<T>;

// All the ways in which a same-shape *UnboxedDto differs from a *Dto:
// - NonEmptyArray -> Array
// - Opaque<T, K> -> T
// - undefined -> null
export type UnboxedDto<T> = DeepUndefinedToNull<DeepEmptiableArray<DeepTransparent<T>>>;

export type DeepOmitModify<T> =
  | {
      [K in keyof T]: T[K] extends never ? any : T[K] extends object ? DeepOmitModify<T[K]> : never;
    }
  | Array<DeepOmitModify<T>>
  | Promise<DeepOmitModify<T>>
  | Set<DeepOmitModify<T>>
  | ReadonlySet<DeepOmitModify<T>>
  | WeakSet<DeepOmitModify<T>>
  | Map<any, DeepOmitModify<T>>
  | WeakMap<any, DeepOmitModify<T>>;

export type DeepUnionPrimitives<U, T> = T extends Builtin
  ? T | U
  : T extends Map<infer K, infer V>
  ? Map<DeepUnionPrimitives<U, K>, DeepUnionPrimitives<U, V>>
  : T extends ReadonlyMap<infer K, infer V>
  ? ReadonlyMap<DeepUnionPrimitives<U, K>, DeepUnionPrimitives<U, V>>
  : T extends WeakMap<infer K, infer V>
  ? WeakMap<DeepUnionPrimitives<U, K>, DeepUnionPrimitives<U, V>>
  : T extends Set<infer V>
  ? Set<DeepUnionPrimitives<U, V>>
  : T extends ReadonlySet<infer V>
  ? ReadonlySet<DeepUnionPrimitives<U, V>>
  : T extends WeakSet<infer V>
  ? WeakSet<DeepUnionPrimitives<U, V>>
  : T extends Promise<infer V>
  ? Promise<DeepUnionPrimitives<U, V>>
  : T extends {}
  ? {
      [K in keyof T]: DeepUnionPrimitives<U, T[K]>;
    }
  : T | U;
